""" adapters_converters_tutorial.py
Python source code for the tutorial on Adapters and Converters.

author: Alberto Semat, tutorial writer for wellrs.com
version: 1.0.0 [2019-04-06]
"""

import sqlite3
import random
import datetime
import csv
import string


# ========================= Per-value Transformations =========================

def test_text_factory():
    """ Test on the `text_factory` converter. """
    
    def uppercase(b):
        print('converting %s' % b)
        return str(b.upper(), encoding='utf-8')
    
    with sqlite3.connect(':memory:') as conn:
        conn.text_factory = uppercase
        conn.execute('CREATE TABLE test(misc TEXT);')
        conn.executemany('INSERT INTO test VALUES(?);',
            ((None, ), (12, ), (34.56, ), 
            ('text-data', ), (b'blob-data', )))
            
        for value, ctype in conn.execute(
            'SELECT misc, typeof(misc) FROM test;').fetchall():
            print('%s is of type %s' % (value, ctype))

# ================================== Adapters =================================

class Coordinate3d:
    """ A 3D Coordinate. """
    
    def __init__(self, x, y, z):
        self.x, self.y, self.z = x, y, z
        
    @staticmethod
    def random_coord():
        return Coordinate3d(
            random.uniform(0, 100),
            random.uniform(0, 100),
            random.uniform(0, 100))
    
    def __conform__(self, protocol):
        if protocol is sqlite3.PrepareProtocol:
            return '(%.3f, %.3f, %.3f)' % (self.x, self.y, self.z)


def test_conform_adapter():
    """ Adapting a class with a `__conform__()` method. """
    with sqlite3.connect(':memory:') as conn:
        conn.execute('CREATE TABLE coords(c);')
        conn.executemany(
            'INSERT INTO coords VALUES (?);', 
            ((Coordinate3d.random_coord(), ) for i in range(10)))
                    
        for c in conn.execute('SELECT c FROM coords;').fetchall():
            print(c[0])
            
            
class PostalAddress:
    
    def __init__(self, recipient, address, city, state):
        self.recipient = recipient
        self.address = address
        self.city, self.state = city, state
        

def adapt_postal_address(address):
    return '%s\n%s\n%s %s' % (address.recipient.upper(), 
        address.address.upper(), 
        address.city.upper(), address.state.upper())

def test_registered_adapter():
    """ Adapting via registered adapters. """
    sqlite3.register_adapter(PostalAddress, adapt_postal_address)
    
    with sqlite3.connect(':memory:') as conn, \
        open('messages.csv', newline='', encoding='utf-8') as src:
        conn.execute('CREATE TABLE mail(' \
            'message TEXT, address TEXT);')
            
        conn.executemany('INSERT INTO mail VALUES (?, ?);',
            ((message, PostalAddress(recipient, address, city, state)) \
            for message, recipient, address, city, state \
            in csv.reader(src)))
        
        for message, address in conn.execute(
            'SELECT * FROM mail;').fetchall():
            print('message: %s\nto: %s\n' % (message, address))
            
       
class Point2d:
    
    def __init__(self, x, y):
        self.x, self.y = x, y
        
    def __repr__(self):
        return 'Point2d(%.3f, %.3f)' % (self.x, self.y)
        return 'Point2d(%.3f, %.3f)' % (self.x, self.y)
    
    def midpoint(self, p):
        return Point2d((self.x + p.x)/2, (self.y + p.y)/2)
        
        
        
class Point2dCheck(Point2d):
    
    def __init__(self, x, y, upper):
        if x > upper or y > upper:
            raise ValueError('Invalid point.')
            
        super().__init__(x, y)
   
        
def adapt_point(p):
    return '(%.3f, %.3f)' % (p.x, p.y)


def test_adapt_subclass():
    """ Adapting an object from a subclass. """
    sqlite3.register_adapter(Point2d, adapt_point)
    sqlite3.register_adapter(Point2dCheck, adapt_point)
    
    with sqlite3.connect(':memory:') as conn:
        conn.execute('CREATE TABLE tab(c TEXT);')
        conn.execute('INSERT INTO tab VALUES(?);',
            (Point2d(1.2, 3.4), ))
        conn.execute('INSERT INTO tab VALUES(?);',
            (Point2dCheck(5.6, 7.8, 10), ))
        
        for p in conn.execute('SELECT * FROM tab;').fetchall():
            print(p[0])


class EMailAddress:
    """ An e-mail address. """
    
    def __init__(self, username, domain):
        """ Creates an e-mail from a user name and a domain. """
        self.username, self.domain = username, domain
        
    def __conform__(self, protocol):
        """ Transforms an `EMailAddress` object into a `str`. """
        if protocol is sqlite3.PrepareProtocol:
            return '%s@%s' % (self.username, self.domain)
        
def adapt_email_address(address):
    """ Transforms an `EMailAddress` object into a `str`. """
    return '%s@%s' % (
        address.username.upper(), 
        address.domain.upper())
        
def test_adapters_precedenece():
    """ Tests the precedence rules for adapters. """
    sqlite3.register_adapter(EMailAddress, adapt_email_address)
    
    with sqlite3.connect(':memory:') as connection:
        connection.execute('CREATE TABLE tab(address);')
        connection.execute(
            'INSERT INTO tab VALUES (?);', 
            (EMailAddress('john.smith', 'gmail.com'), ))
        curs = connection.execute(
            'SELECT address FROM tab;')
        print(curs.fetchone()[0])
        
        
def test_adapters():
    """ Performs all tests on adapters. """
    test_sep('Adapting via the `__conform__` method')
    test_conform_adapter()
    test_sep('Adapting via a registered adapter')
    test_registered_adapter()
    test_sep('Adapting a Subclass')
    test_adapt_subclass()
    test_sep('Adapters Precedence')
    test_adapters_precedenece()
            

# ================================= Converters ================================

def generate_lines(n=10, lower=-100, upper=100):
    """ Generates `n` lines within certain boundaries. """
    
    def generate_point(lower, upper):
        """ Generates a single point. """
        return Point2d(random.uniform(lower, upper),
            random.uniform(lower, upper))
    
    for p in range(n):
        yield (generate_point(lower, upper), 
            generate_point(lower, upper))
            
def convert_point(p):
    x, y = p[1 : -1].split(b',')
    return Point2d(float(x), float(y))

def test_converters_by_decltype():
    """ Converters matched by declared type. """
    sqlite3.register_adapter(Point2d, adapt_point)
    sqlite3.register_converter('point', convert_point)
    
    with sqlite3.connect(
        ':memory:',
        detect_types=sqlite3.PARSE_DECLTYPES) as conn:
        conn.execute('CREATE TABLE lines(' \
            'start POINT, end POINT);')
        
        conn.executemany('INSERT INTO lines VALUES(?, ?);',
            generate_lines())
            
        for start, stop in conn.execute(
            'SELECT * FROM lines;').fetchall():
            print('line from %s to %s' % (start, stop))
            
def test_converters_by_colname():
    """ Converters matched by column names. """
    sqlite3.register_adapter(Point2d, adapt_point)
    sqlite3.register_converter('point', convert_point)
    
    with sqlite3.connect(
        ':memory:',
        detect_types=sqlite3.PARSE_COLNAMES) as conn:
        conn.execute('CREATE TABLE lines(start, end);')
        
        conn.executemany('INSERT INTO lines VALUES(?, ?);',
            generate_lines())
            
        for start, stop in conn.execute(
            'SELECT start AS "s [point]", end AS "e [point]" '\
            'FROM lines;').fetchall():
            print('line from %s to %s' % (start, stop))
        
def words_generator(n=20, minlen=5, maxlen=10):
    """ Generates `n` words. """
    for i in range(n):
        word = ''
        for i in range(random.randint(minlen, maxlen)):
            word += random.choice(string.ascii_letters)
            
        yield (word, )
        
def test_converters_precedence():
    """ Tests the precedence rules for converters. """
    def lower(s):
        return str(s.lower(), encoding='utf-8')
        
    def upper(s):
        return str(s.upper(), encoding='utf-8')
    
    sqlite3.register_converter('text', lower)
    sqlite3.register_converter('word', upper)
    
    with sqlite3.connect(
        ':memory:',
        detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES) \
        as conn:        
        conn.execute('CREATE TABLE test(word TEXT);')
            
        conn.executemany('INSERT INTO test VALUES (?);', words_generator(10))
        
        print('Using PARSE_COLNAMES: upper'.center(50, '.'))
        for word in conn.execute(
            'SELECT word AS "w [word]" FROM test;').fetchall():
            print(word[0])
            
        print('Using PARSE_DECLTYPES: lower'.center(50, '.'))
        for word in conn.execute(
            'SELECT * FROM test;').fetchall():
            print(word[0])
                
def test_textfactory_precedence():
    """ `text_factory` has lower precendece than other converters. """
    def lower(s):
        return str(s.lower(), encoding='utf-8')
        
    def upper(s):
        return str(s.upper(), encoding='utf-8')
        
    def titlecase(s):
        return str(s.title(), encoding='utf-8')
        
    # Delete a previously registered converter
    sqlite3.register_converter('text', None)
    
    with sqlite3.connect(
        ':memory:',
        detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES) \
        as conn:
        conn.text_factory = titlecase
            
        conn.execute('CREATE TABLE test(word TEXT);')
            
        conn.executemany('INSERT INTO test VALUES (?);', words_generator(5))
                
        print('Without any converter: row_factory'.center(50, '.'))
        for word in conn.execute(
            'SELECT * FROM test;').fetchall():
            print(word[0])
                    
        sqlite3.register_converter('text', lower)
        sqlite3.register_converter('word', upper)
        
        print('Using PARSE_COLNAMES: upper'.center(50, '.'))
        for word in conn.execute(
            'SELECT word AS "w [word]" FROM test;').fetchall():
            print(word[0])
            
        print('Using PARSE_DECLTYPES: lower'.center(50, '.'))
        for word in conn.execute(
            'SELECT * FROM test;').fetchall():
            print(word[0])
            
def test_converters():
    """ Performs all tests on converters. """
    test_sep('Converters by Declared Type')
    test_converters_by_decltype()
    test_sep('Converters by Column Name')
    test_converters_by_colname()
    test_sep('Converters\' Precedence')
    test_converters_precedence()
    test_sep('Precedence over `text_factory`')
    test_textfactory_precedence()
    
        
# ============================== Misc ========================================

def test_null_values():
    """ No converter or adapter is called for NULL values. """
    
    def convert_misc(b):
        print('converting a MISC value: %s' % b)
        return str(b, encoding='utf-8')
        
    def adapt_null(s):
        print('Adapting a None value.')
        return 'NULL'
        
    def adapt_str(s):
        print('Adapting a str value: %s' % s)
        return s
    
    sqlite3.register_adapter(None, adapt_null)
    sqlite3.register_adapter(str, adapt_str)
    sqlite3.register_converter('misc', convert_misc)
    
    with sqlite3.connect(':memory:',
        detect_types=sqlite3.PARSE_DECLTYPES) as conn:        
        conn.execute(
            'CREATE TABLE test(c misc);')
                
        conn.executemany(
            'INSERT INTO test VALUES (?);',
            (('a', ), (None, ), ('b', )))
                
        for value, ctype in conn.execute(
            'SELECT c, typeof(c) FROM test;').fetchall():
            print('"%s" is of type %s' % (value, ctype.upper()))
        
def test_date_by_decltypes():
    """ Tests default date adapters and converters. """
    with sqlite3.connect(
        ':memory:',
        detect_types=sqlite3.PARSE_DECLTYPES) as conn:
        conn.execute(
            'CREATE TABLE test(' \
            'd DATE,' \
            'ts TIMESTAMP);')
            
        today = datetime.date.today()
        now = datetime.datetime.now()
        
        conn.execute(
            'INSERT INTO test(d, ts) ' \
            'VALUES (?, ?);', (today, now))
            
        date, type_1, timestamp, type_2,  = conn.execute(
            'SELECT d, typeof(d), ' \
            'ts, typeof(ts) FROM test;').fetchone()
        print('today is %s (from %s to %s)\n' \
            'now is %s (from %s to %s)\n' \
            % (date, type_1.upper(), type(date),
            timestamp, type_2.upper(), type(timestamp)))
        
def test_date_by_colnames():
    """ Match default date converters by column names. """
    with sqlite3.connect(
        ':memory:',
        detect_types=sqlite3.PARSE_COLNAMES) as conn:
        date, type_1, timestamp, type_2 = conn.execute(
            'SELECT CURRENT_DATE AS "d [date]", typeof(CURRENT_DATE), ' \
            'CURRENT_TIMESTAMP AS "ts [timestamp]", ' \
            'typeof(CURRENT_TIMESTAMP);').fetchone()
        print('current date is %s (from %s to %s)\n' \
            'current timestamp is %s (from %s to %s)\n' \
            % (date, type_1.upper(), type(date),
            timestamp, type_2.upper(), type(timestamp)))
            

def test_misc():
    """ Tests on miscellaneous topics. """
    test_sep('Beware NULL values')
    test_null_values()
    test_sep('Default Date Adapters & Converters')
    test_date_by_decltypes()
    test_date_by_colnames()

# =================================== Test ====================================


def test_sep(title, sep='-'):
    print('\n' + title.center(50, sep))
    
def test_all():
    """ Performs all test in the tutorial. """
    test_sep('PER-VALUE TRANSFORMATIONS', '*')
    test_text_factory()
    test_sep('ADAPTERS', '*')
    test_adapters()
    test_sep('CONVERTERS', '*')
    test_converters()
    test_sep('MISC', '*')
    test_misc()
    

test_all()

